/*  This is a -*- C++ -*- header file.

    Copyright (C) 2019-2021 Julian Rüth

    This file is part of e-antic

    e-antic is free software: you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (LGPL) as published
    by the Free Software Foundation; either version 3.0 of the License, or
    (at your option) any later version.  See <https://www.gnu.org/licenses/>.
*/

#ifndef E_ANTIC_FORWARD_HPP
#define E_ANTIC_FORWARD_HPP

// This file contains forward declarations for all the C++ classes defined by
// e-antic to speed up compilation when included in header files.

namespace eantic {

class renf_elem_class;
class renf_class;

}

namespace cereal {

class access;

}

#endif

