/*  This is a -*- C++ -*- header file.

    Copyright (C) 2021 Julian Rüth

    This file is part of e-antic

    e-antic is free software: you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License (LGPL) as published
    by the Free Software Foundation; either version 3.0 of the License, or
    (at your option) any later version.  See <https://www.gnu.org/licenses/>.
*/

#ifndef E_ANTIC_E_ANTIC_HPP
#define E_ANTIC_E_ANTIC_HPP

// Include this header to use the full C++ interface of e-antic.
// For better build times, consider including only forward.hpp or
// some of the following headers instead.

#include "renf_class.hpp"
#include "renf_elem_class.hpp"

#endif
